/**
 * This file is part of GenomeView, a genome browser and annotation curator
 * 
 * Copyright (C) 2007-2013 Thomas Abeel and contributors
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 * Project: http://genomeview.org/
 */
package net.sf.genomeview.core;

import java.util.HashMap;
import java.util.Map;

import be.abeel.io.LineIterator;

/**
 * Class to get the database information for a dbxref.
 * 
 * @author Thomas Abeel
 * 
 */
public class Dbxref {

    static {
        load();
    }

    /* Maps abbrev to dbURL */
    private static Map<String, String> dbxrefmap;

    private static void load() {
        dbxrefmap = new HashMap<String, String>();
        LineIterator it = new LineIterator(Dbxref.class.getResourceAsStream("/conf/dbxref.txt"));
        it.setSkipComments(true);
        it.setSkipBlanks(true);
        String abb = null;
        String url = null;
        for (String line : it) {
            if (line.startsWith("Abbrev:")) {
                abb = line.split(": ")[1];
            }
            if (line.startsWith("Db_URL:")) {
                url = line.split(": ")[1];
                dbxrefmap.put(abb, "http://" + url);
            }
        }

    }
    public static String getURL(String abbreviation){
        System.out.println("dbxref query: "+abbreviation);
        if(dbxrefmap.containsKey(abbreviation))
            return dbxrefmap.get(abbreviation);
        return null;
    }

}
