/**
 * This file is part of GenomeView, a genome browser and annotation curator
 * 
 * Copyright (C) 2007-2013 Thomas Abeel and contributors
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 * Project: http://genomeview.org/
 */
package net.sf.genomeview.gui.search;

import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;

import net.sf.genomeview.data.Model;
import net.sf.genomeview.gui.search.SearchDialog.SequenceType;
import net.sf.jannot.Location;
import be.abeel.gui.TitledComponent;

/**
 * 
 * @author Thomas Abeel
 * 
 */
class MotifSearchPane extends SearchPanel {

	private static final long serialVersionUID = -3270709193426284702L;

	MotifSearchPane(final Model model) {
		gc.fill = GridBagConstraints.BOTH;
	
		
		final JTextField seq = new JTextField(40);
		this.setFocusField(seq);
		
		JButton search = new JButton("Search");

		final MotifSearchResultModel mrm = new MotifSearchResultModel(model);
		final JTable results = new JTable(mrm);
		results.addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent e) {
				int row = results.getSelectedRow();
				if (row >= 0) {
					Location l = mrm.getLocation(row);
					model.vlm.center((l.start() + l.end()) / 2);
				}
			}
		});
		final JComboBox type = new JComboBox(SequenceType.values());

		search.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				mrm.clear();
				model.clearHighlights();
				mrm.search(model, seq.getText().trim(), 
						(SequenceType) type.getSelectedItem());

			}

		});
		seq.addKeyListener(new KeyAdapter() {

			/*
			 * (non-Javadoc)
			 * 
			 * @see java.awt.event.KeyAdapter#keyTyped(java.awt.event.KeyEvent)
			 */
			@Override
			public void keyReleased(KeyEvent e) {

				int key = e.getKeyCode();
				if (key == KeyEvent.VK_ENTER) {
					mrm.clear();
					model.clearHighlights();
					mrm.search(model, seq.getText().trim(),
							(SequenceType) type.getSelectedItem());
				} else {
					super.keyTyped(e);
				}

			}

		});
		

		
		/* Query sequence box */
		gc.weightx = 1;
		gc.gridwidth = 4;
		gc.weighty = 0.1;
		add(new TitledComponent("Query sequence (N for ambiguous nucletodie, X for ambiguous amino acid)", seq), gc);
		gc.weighty = 0;
		gc.gridwidth = 1;
		gc.gridy++;
		add(type, gc);
		gc.weightx = 0;
		gc.gridx++;
		add(search, gc);
	

		/* Result table */
		gc.gridx = 0;
		gc.gridwidth = 4;
		gc.gridy++;
		gc.weighty = 1;
		gc.weightx = 1;
		add(new TitledComponent("Result locations", new JScrollPane(results)), gc);

	}

}